require 'json'

module UU
  module OS
    module REST

      # Each DTO should define constructor accepting JSON, +Hash+ or instance of same DTO.
      # For purpose of serialization, methods +to_hash+ and +to_json+ should be defined.
      # This ancestor provides all this functionality so they need to be overriden
      # only in special cases.
      #
      # Descendants of this class must define getters and setters (which are delegating to
      # +attributes+ from this class). These setters and getters are responsible for any
      # necessary data transformations.
      class DTO

        # Creates new instance of DTO. In case there is some problem with default parsing
        # of +init_data+, method +handle_custom_init_data+ is invoked in order to provide
        # descendants option to handle special types of +init_data+.
        #
        # @param [Object] init_data Data used for instantiation. Accepted values:
        #   * JSON string with serialized DTO
        #   * Hash containing DTO attributes
        #   * Instance of same DTO
        def initialize(init_data = nil)
          @attributes = {}
          original_data = init_data
          if init_data.nil?
            return
          elsif init_data.kind_of?self.class
            # In case of self instance only reuse existing DTO attributes
            @attributes = init_data.send(:attributes)
            return
          elsif init_data.kind_of?String
            # Input of type String should be JSON with serialized DTO
            begin
              init_data = JSON.parse(init_data, :symbolize_names => true)
            rescue Exception => e
              begin
                handle_custom_init_data(original_data, e)
                return
              rescue Exception => e1
                if init_data.empty?
                  # In case init_data is empty String (invalid JSON) and init_data are not
                  # handled by particular DTO type, consider empty init_data as nil init_data
                  # and create empty instance of this DTO.
                  return
                else
                  raise e1
                end
              end
            end
          end
          if !init_data.kind_of?Hash
            error = ArgumentError.new("Parameter must be a valid JSON String, a Hash or an instance of #{self.class}, but was #{init_data.class}.")
            handle_custom_init_data(original_data, error)
            return
          end
          # Handle attributes given as Hash (either given directly or deserialized from JSON)
          init_data.entries.each do |attribute|
            # Delegate each attribute to setter in particular DTO
            setter = get_setter(attribute[0])
            begin
              self.send(setter, attribute[1])
            rescue Exception => e
              handle_custom_init_data(original_data, e)
              return
            end
          end
        end

        # Serializes DTO to Hash. Serialization is done recursively, that means
        # +to_hash+ is invoked on all attributes of type DTO. Manipulating data
        # in returned structure does not change data in original DTOs.
        # This method should be overridden in descendants in case some additional
        # attributes transformation is necessary.
        #
        # @return [Hash] Hash representing DTO
        def to_hash
          result = {}
          @attributes.entries.each do |attribute|
            if attribute[1].kind_of?UU::OS::REST::DTO
              result[attribute[0]] = attribute[1].to_hash
            else
              result[attribute[0]] = attribute[1]
            end
          end
          result
        end

        # Serialize DTO to JSON. DTO is first serialized to Hash using method
        # +to_hash+, result is than serialized to JSON.
        # This method should be overridden in descendants in case some additional
        # attributes transformation is necessary.
        #
        # @param [Hash] options Serialization options
        #
        # @return [String] JSON with serialized DTO
        def to_json(options = nil)
          to_hash.to_json(options)
        end

        # Handle invocation of missing setter while constructing new DTO.
        #
        # @private
        def method_missing(method, *args, &block)
          if (method.to_s[/=$/]) && (args.size == 1)
            UU::OS::Logger.new(self).warning("#{self.class} does not contain setter \"#{method}\". Either DTO is being created with wrong parameters or there is some change in API.")
          else
            super.method_missing(method, args, &block)
          end
        end

        # Alias default to_s method to be able to use it in inspect even
        # if to_s method is overriden in DTO descendants.
        #
        # @private
        alias_method :default_to_s, :to_s

        # @private
        def to_s
          str = "{ \"#{self.class.name}\" : { "
          attributes.entries.each do |attribute|
            if (attribute[1].kind_of?UU::OS::REST::DTO) && (!attribute[1].kind_of?UU::OS::UESURI)
              str << "\"#{attribute[0]}\" : #{attribute[1].to_s} , "
            else
              str << "\"#{attribute[0]}\" : \"#{attribute[1].to_s}\" , "
            end
          end
          str = str[0..-4] if attributes.size > 0
          str << ' } }'
          return str
        end

        # @private
        def inspect
          # Override default inspect to suppress printing of
          # all inner attributes of DTO.
          default_to_s
        end

        protected

        # Gets all DTO attributes. Descendants of this class
        # should use this method for implementation of
        # getters and setters.
        #
        # @return [Hash] Hash with DTO attributes
        def attributes
          @attributes
        end

        # Sets DTO attributes.
        #
        # @param [Hash] val Hash representing DTO attributes
        def attributes=(val)
          @attributes = val
        end

        # Returns expected setter method for given attribute (camel
        # case names are transformed to snake case).
        # Descendants may override this method in case naming convention
        # is not enough to create valid setter from attribute (e.g. when
        # attribute name differs on REST API and Ruby API).
        #
        # @param [Symbol] attribute Attribute for which to get setter method
        #
        # @return [Symbol] Setter method
        def get_setter(attribute)
          # Based on ActiveSupport::Inflector.underscore
          result = attribute.to_s
          result.gsub!(/([A-Z]+)([A-Z][a-z])/,'\1_\2')
          result.gsub!(/([a-z\d])([A-Z])/,'\1_\2')
          result.tr!('-', '_')
          (result.downcase << '=').to_sym
        end

        # Some DTOs may accept some other data than Hash, JSON or selfinstance.
        # This method allows to implement custom DTO initialization.
        #
        # @param [Object] init_data Custom init data.
        # @param [Exception] error Error which occured by standard data processing
        def handle_custom_init_data(init_data, error)
          raise error
        end

        private

        # @private
        def type=(val)
          # DTOs containing internal attribute with some other DTO using generics are serialized to JSON
          # so that internal DTOs containing attribute type with DTO class name. But this attribute is
          # not declared on DTO itself and can be safely dumped.
        end

      end

    end
  end
end
